"""
A set of helper utilities to utilize taskgroups in better ways.
"""

import asyncio

from . import PersistentTaskGroup

__all__ = ("as_completed_safe",)


async def as_completed_safe(coros, timeout=None):
    """
    This is a safer version of :func:`asyncio.as_completed()` which uses
    :class:`PersistentTaskGroup` as an underlying coroutine lifecycle keeper.

    Upon a timeout, it raises :class:`asyncio.TimeoutError` immediately
    and cancels all remaining tasks or coroutines.

    This requires Python 3.11 or higher to work properly with timeouts.

    .. versionadded:: 1.6
    """
    async with PersistentTaskGroup() as tg:
        tasks = []
        for coro in coros:
            t = tg.create_task(coro)
            tasks.append(t)
        await asyncio.sleep(0)
        try:
            for result in asyncio.as_completed(tasks, timeout=timeout):
                yield result
        except GeneratorExit:
            # This happens when as_completed() is timeout.
            raise asyncio.TimeoutError()



from importlib.metadata import version

# Import submodules only when installed properly
from . import (
    context,
    func,
    server,
    taskgroup,
    timer,
)
from . import defer as _defer
from . import fork as _fork
from . import iter as _iter

__all__ = (
    *context.__all__,
    *_defer.__all__,
    *_fork.__all__,
    *func.__all__,
    *_iter.__all__,
    *server.__all__,
    *taskgroup.__all__,
    *timer.__all__,
    "__version__",
    # "@aiotools.server" still works,
    # but server_context is provided to silence typecheckers.
    "main_context",
    "server_context",
)

from .context import *  # noqa
from .defer import *  # noqa
from .fork import *  # noqa
from .func import *  # noqa
from .iter import *  # noqa
from .server import *  # noqa
from .taskgroup import *  # noqa
from .timer import *  # noqa

main_context = server._main_ctxmgr
server_context = server._server_ctxmgr

__version__ = version("aiotools")



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# This file is included in the aiotools open source project
# given that EdgeDB's Apache 2.0 license is compatible with
# aiotools' MIT license.
# The original source code is taken from:
# https://github.com/edgedb/edgedb-python/blob/bcbe005/edgedb/_taskgroup.py

import asyncio
import itertools
from contextvars import ContextVar

from ..compat import current_task, get_running_loop
from .common import create_task_with_name, patch_task
from .types import TaskGroupError

__all__ = [
    "TaskGroup",
]


current_taskgroup: ContextVar["TaskGroup"] = ContextVar("current_taskgroup")
__all__.append("current_taskgroup")


class TaskGroup:
    def __init__(self, *, name=None):
        if name is None:
            self._name = f"tg-{_name_counter()}"
        else:
            self._name = str(name)

        self._entered = False
        self._exiting = False
        self._aborting = False
        self._loop = get_running_loop()
        self._parent_task = None
        self._parent_cancel_requested = False
        self._tasks = set()
        self._unfinished_tasks = 0
        self._errors = []
        self._base_error = None
        self._on_completed_fut = None

    def get_name(self):
        return self._name

    def __repr__(self):
        msg = f"<TaskGroup {self._name!r}"
        if self._tasks:
            msg += f" tasks:{len(self._tasks)}"
        if self._unfinished_tasks:
            msg += f" unfinished:{self._unfinished_tasks}"
        if self._errors:
            msg += f" errors:{len(self._errors)}"
        if self._aborting:
            msg += " cancelling"
        elif self._entered:
            msg += " entered"
        msg += ">"
        return msg

    async def __aenter__(self):
        if self._entered:
            raise RuntimeError(f"TaskGroup {self!r} has been already entered")
        self._entered = True

        self._parent_task = current_task()

        if self._parent_task is None:
            raise RuntimeError(f"TaskGroup {self!r} cannot determine the parent task")
        patch_task(self._parent_task)
        self._current_taskgroup_token = current_taskgroup.set(self)
        return self

    async def __aexit__(self, et, exc, tb):
        self._exiting = True
        propagate_cancelation = False

        if exc is not None and self._is_base_error(exc) and self._base_error is None:
            self._base_error = exc

        if et is asyncio.CancelledError:
            if self._parent_cancel_requested:
                # Only if we did request task to cancel ourselves
                # we mark it as no longer cancelled.
                self._parent_task.__cancel_requested__ = False
            else:
                propagate_cancelation = True

        if et is not None and not self._aborting:
            # Our parent task is being cancelled:
            #
            #    async with TaskGroup() as g:
            #        g.create_task(...)
            #        await ...  # <- CancelledError
            #
            if et is asyncio.CancelledError:
                propagate_cancelation = True

            # or there's an exception in "async with":
            #
            #    async with TaskGroup() as g:
            #        g.create_task(...)
            #        1 / 0
            #
            self._abort()

        # We use while-loop here because "self._on_completed_fut"
        # can be cancelled multiple times if our parent task
        # is being cancelled repeatedly (or even once, when
        # our own cancellation is already in progress)
        while self._unfinished_tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = self._loop.create_future()

            try:
                await self._on_completed_fut
            except asyncio.CancelledError:
                if not self._aborting:
                    # Our parent task is being cancelled:
                    #
                    #    async def wrapper():
                    #        async with TaskGroup() as g:
                    #            g.create_task(foo)
                    #
                    # "wrapper" is being cancelled while "foo" is
                    # still running.
                    propagate_cancelation = True
                    self._abort()

            self._on_completed_fut = None

        assert self._unfinished_tasks == 0
        self._on_completed_fut = None  # no longer needed
        current_taskgroup.reset(self._current_taskgroup_token)

        if self._base_error is not None:
            raise self._base_error

        if propagate_cancelation:
            # The wrapping task was cancelled; since we're done with
            # closing all child tasks, just propagate the cancellation
            # request now.
            raise asyncio.CancelledError()

        if et is not None and et is not asyncio.CancelledError:
            self._errors.append(exc)

        if self._errors:
            # Exceptions are heavy objects that can have object
            # cycles (bad for GC); let's not keep a reference to
            # a bunch of them.
            errors = self._errors
            self._errors = None

            me = TaskGroupError("unhandled errors in a TaskGroup", errors)
            raise me from None

    def create_task(self, coro, *, name=None):
        if not self._entered:
            raise RuntimeError(f"TaskGroup {self!r} has not been entered")
        if self._exiting and self._unfinished_tasks == 0:
            raise RuntimeError(f"TaskGroup {self!r} is awaiting in exit")
        task = create_task_with_name(coro, name=name)
        task.add_done_callback(self._on_task_done)
        self._unfinished_tasks += 1
        self._tasks.add(task)
        return task

    def _is_base_error(self, exc):
        assert isinstance(exc, BaseException)
        return isinstance(exc, (SystemExit, KeyboardInterrupt))

    def _abort(self):
        self._aborting = True

        for t in self._tasks:
            if not t.done():
                t.cancel()

    def _on_task_done(self, task):
        try:
            self._unfinished_tasks -= 1
            assert self._unfinished_tasks >= 0

            if self._exiting and not self._unfinished_tasks:
                if not self._on_completed_fut.done():
                    self._on_completed_fut.set_result(True)

            if task.cancelled():
                return

            exc = task.exception()
            if exc is None:
                return

            self._errors.append(exc)
            if self._is_base_error(exc) and self._base_error is None:
                self._base_error = exc

            if self._parent_task.done():
                # Not sure if this case is possible, but we want to handle
                # it anyways.
                self._loop.call_exception_handler({
                    "message": (
                        f"Task {task!r} has errored out but its parent "
                        f"task {self._parent_task} is already completed"
                    ),
                    "exception": exc,
                    "task": task,
                })
                return
        finally:
            self._tasks.discard(task)

        self._abort()
        if not self._parent_task.__cancel_requested__:
            # If parent task *is not* being cancelled, it means that we want
            # to manually cancel it to abort whatever is being run right now
            # in the TaskGroup.  But we want to mark parent task as
            # "not cancelled" later in __aexit__.  Example situation that
            # we need to handle:
            #
            #    async def foo():
            #        try:
            #            async with TaskGroup() as g:
            #                g.create_task(crash_soon())
            #                await something  # <- this needs to be canceled
            #                                 #    by the TaskGroup, e.g.
            #                                 #    foo() needs to be cancelled
            #        except Exception:
            #            # Ignore any exceptions raised in the TaskGroup
            #            pass
            #        await something_else     # this line has to be called
            #                                 # after TaskGroup is finished.
            self._parent_cancel_requested = True
            self._parent_task.cancel()


_name_counter = itertools.count(1).__next__



"""
Provides an implementation of asynchronous context manager and its applications.

.. versionchanged::

   As aiotools 1.7+ drops support for Python 3.7 or earlier, most attributes
   are just aliases of the standard contextlib.
"""

import asyncio
import contextlib
from contextvars import ContextVar
from typing import (
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
)

from .types import AClosable, AsyncClosable

__all__ = [
    "resetting",
    "AsyncContextManager",
    "async_ctx_manager",
    "actxmgr",
    "aclosing",
    "closing_async",
    "AsyncContextGroup",
    "actxgroup",
]

T = TypeVar("T")
T_AClosable = TypeVar("T_AClosable", bound=AClosable)
T_AsyncClosable = TypeVar("T_AsyncClosable", bound=AsyncClosable)


AbstractAsyncContextManager = contextlib.AbstractAsyncContextManager
AsyncContextManager = contextlib._AsyncGeneratorContextManager
AsyncExitStack = contextlib.AsyncExitStack
async_ctx_manager = contextlib.asynccontextmanager


class resetting(Generic[T]):
    """
    An extra context manager to auto-reset the given context variable.
    It supports both the standard contextmanager protocol and the
    async-contextmanager protocol.

    .. versionadded:: 1.8.0
    """

    def __init__(self, ctxvar: ContextVar[T], value: T) -> None:
        self._ctxvar = ctxvar
        self._value = value

    def __enter__(self) -> None:
        self._token = self._ctxvar.set(self._value)

    async def __aenter__(self) -> None:
        self._token = self._ctxvar.set(self._value)

    def __exit__(self, *exc_info) -> Optional[bool]:
        self._ctxvar.reset(self._token)
        return None

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        self._ctxvar.reset(self._token)
        return None


class aclosing(Generic[T_AClosable]):
    """
    An analogy to :func:`contextlib.closing` for async generators.

    The motivation has been proposed by:

    * https://github.com/njsmith/async_generator
    * https://vorpus.org/blog/some-thoughts-on-asynchronous-api-design-\
in-a-post-asyncawait-world/#cleanup-in-generators-and-async-generators
    * https://www.python.org/dev/peps/pep-0533/
    """

    def __init__(self, thing: T_AClosable) -> None:
        self.thing = thing

    async def __aenter__(self) -> T_AClosable:
        return self.thing

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        await self.thing.aclose()
        return None


class closing_async(Generic[T_AsyncClosable]):
    """
    An analogy to :func:`contextlib.closing` for objects defining the ``close()``
    method as an async function.

    .. versionadded:: 1.5.6
    """

    def __init__(self, thing: T_AsyncClosable) -> None:
        self.thing = thing

    async def __aenter__(self) -> T_AsyncClosable:
        return self.thing

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        await self.thing.close()
        return None


class AsyncContextGroup:
    """
    Merges a group of context managers into a single context manager.
    Internally it uses :func:`asyncio.gather()` to execute them with overlapping,
    to reduce the execution time via asynchrony.

    Upon entering, you can get values produced by the entering steps from
    the passed context managers (those ``yield``-ed) using an ``as`` clause of
    the ``async with``
    statement.

    After exits, you can check if the context managers have finished
    successfully by ensuring that the return values of ``exit_states()`` method
    are ``None``.

    .. note::

       You cannot return values in context managers because they are
       generators.

    If an exception is raised before the ``yield`` statement of an async
    context manager, it is stored at the corresponding manager index in the
    as-clause variable.  Similarly, if an exception is raised after the
    ``yield`` statement of an async context manager, it is stored at the
    corresponding manager index in the ``exit_states()`` return value.

    Any exceptions in a specific context manager does not interrupt others;
    this semantic is same to ``asyncio.gather()``'s when
    ``return_exceptions=True``.  This means that, it is user's responsibility
    to check if the returned context values are exceptions or the intended ones
    inside the context body after entering.

    :param context_managers: An iterable of async context managers.
                             If this is ``None``, you may add async context
                             managers one by one using the :meth:`~.add`
                             method.

    """

    def __init__(
        self, context_managers: Optional[Iterable[AbstractAsyncContextManager]] = None
    ):  # noqa
        self._cm = list(context_managers) if context_managers else []
        self._cm_yields: List[asyncio.Task] = []
        self._cm_exits: List[asyncio.Task] = []

    def add(self, cm):
        """
        TODO: fill description
        """
        self._cm.append(cm)

    async def __aenter__(self):
        # Exceptions in context managers are stored into _cm_yields list.
        # NOTE: There is no way to "skip" the context body even if the entering
        #       process fails.
        self._cm_yields[:] = await asyncio.gather(
            *(e.__aenter__() for e in self._cm), return_exceptions=True
        )
        return self._cm_yields

    async def __aexit__(self, *exc_info):
        # Clear references to context variables.
        self._cm_yields.clear()
        # Exceptions are stored into _cm_exits list.
        self._cm_exits[:] = await asyncio.gather(
            *(e.__aexit__(*exc_info) for e in self._cm), return_exceptions=True
        )

    def exit_states(self):
        """
        TODO: fill description
        """
        return self._cm_exits


# Shorter aliases
actxmgr = async_ctx_manager
actxgroup = AsyncContextGroup



from __future__ import annotations

import builtins
import textwrap
import traceback
from collections.abc import Awaitable

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore  # noqa
from types import TracebackType


class AsyncExceptionHandler(Protocol):
    """
    A shorthand for an async exception handler type.
    This is always called under exception context where
    :func:`sys.exc_info()` is available.
    """

    def __call__(
        self,
        exc_type: type[BaseException],
        exc_obj: BaseException,
        exc_tb: TracebackType,
    ) -> Awaitable[None]: ...


if not hasattr(builtins, "ExceptionGroup"):

    class MultiError(Exception):  # type: ignore[no-redef]
        def __init__(self, msg, errors=()):
            if errors:
                types = set(type(e).__name__ for e in errors)
                msg = f"{msg}; {len(errors)} sub errors: ({', '.join(types)})"
                for er in errors:
                    msg += f"\n + {type(er).__name__}: {er}"
                    if er.__traceback__:
                        er_tb = "".join(traceback.format_tb(er.__traceback__))
                        er_tb = textwrap.indent(er_tb, " | ")
                        msg += f"\n{er_tb}\n"
            super().__init__(msg)
            self.__errors__ = tuple(errors)

        def get_error_types(self):
            return {type(e) for e in self.__errors__}

        def __reduce__(self):
            return (type(self), (self.args,), {"__errors__": self.__errors__})

    class TaskGroupError(MultiError):  # type: ignore[no-redef]
        """
        An alias to :exc:`MultiError`.
        """

        pass

else:

    class MultiError(ExceptionGroup):  # type: ignore[no-redef,name-defined]  # noqa: F821
        def __init__(self, msg, errors=()):
            super().__init__(msg, errors)
            self.__errors__ = errors

        def get_error_types(self):
            return {type(e) for e in self.exceptions}

    class TaskGroupError(MultiError):  # type: ignore[no-redef]
        pass



import asyncio

from .types import MultiError, TaskGroupError

__all__ = [
    "MultiError",
    "TaskGroup",
    "TaskGroupError",
    "PersistentTaskGroup",
    "as_completed_safe",
    "persistent",
    "persistent_compat",
    "current_taskgroup",
    "current_ptaskgroup",
    "has_contextvars",
]

if hasattr(asyncio, "TaskGroup"):
    from . import persistent
    from .base import *  # noqa
    from .persistent import *  # noqa
else:
    from . import persistent_compat
    from .base_compat import *  # type: ignore  # noqa
    from .persistent_compat import *  # type: ignore  # noqa

from .utils import as_completed_safe  # noqa

has_contextvars = True  # As of aiotools 1.8.0 and Python 3.9, this is always true.
